@@ -30,134 +30,199 @@
     #include <io.h>
 #endif
 
+/* Use our own getopt implementation. */
+static int opind = 1;
+static int operr = 1;
+static int opopt;
+static char *oparg;
+
+static int getopt_impl(int argc, char * const argv[], const char *optstring) {
+    static int optpos = 1;
+    const char *arg;
+
+    /* Reset? */
+    if (opind == 0) {
+        opind = !!argc;
+        optpos = 1;
+    }
+
+    arg = argv[opind];
+    if (arg && strcmp(arg, "--") == 0) {
+        opind++;
+        return -1;
+    } else if (!arg || arg[0] != '-' || !isalnum(arg[1])) {
+        return -1;
+    } else {
+        const char *opt = strchr(optstring, arg[optpos]);
+        opopt = arg[optpos];
+        if (!opt) {
+            if (operr) {
+                fprintf(stderr, "%s: illegal option: %c\n", argv[0], opopt);
+                exit(1);
+            }
+            return '?';
+        } else if (opt[1] == ':') {
+            if (arg[optpos + 1]) {
+                oparg = (char *)arg + optpos + 1;
+                opind++;
+                optpos = 1;
+                return opopt;
+            } else if (argv[opind + 1]) {
+                oparg = (char *)argv[opind + 1];
+                opind += 2;
+                optpos = 1;
+                return opopt;
+            } else {
+                if (operr) {
+                    fprintf(stderr,
+                            "%s: option requires an argument: %c\n",
+                            argv[0], opopt);
+                    exit(1);
+                }
+                return *optstring == ':' ? ':' : '?';
+            }
+        } else {
+            if (!arg[++optpos]) {
+                opind++;
+                optpos = 1;
+            }
+            return opopt;
+        }
+    }
+}
+
 #include "common.h"
 #include "libbz3.h"
 
-#define MODE_UNSPECIFIED 0
-#define MODE_DECODE -1
+#define MODE_DECODE 0
 #define MODE_ENCODE 1
 #define MODE_TEST 2
 
-int is_dir(const char * path) {
+static void version() {
+    fprintf(stdout,
+            "bzip3 "VERSION"\n"
+            "Copyright (C) by Kamila Szewczyk, 2022.\n"
+            "License: GNU Lesser GPL version 3 <https://www.gnu.org/licenses/lgpl-3.0.en.html>\n");
+    }
+
+static void help() {
+    fprintf(stdout,
+            "Bzip3 - better and stronger spiritual successor to bzip2.\n"
+            "Usage: bzip3 [-e/-d/-t/-c/-h/-V] [-b block_size] [-j jobs] files...\n"
+            "Operations:\n"
+            "  -e: encode\n"
+            "  -d: decode\n"
+            "  -t: test\n"
+            "  -h: help\n"
+            "  -f: force overwrite output if it already exists\n"
+            "  -V: version\n"
+            "Extra flags:\n"
+            "  -c: force reading/writing from standard streams\n"
+            "  -b N: set block size in MiB {16}\n"
+#ifdef PTHREAD
+            "  -j N: set the amount of parallel threads\n"
+#endif
+            "\n"
+            "Report bugs to: https://github.com/kspalaiologos/bzip3\n"
+            "\n");
+}
+
+static int is_dir(const char * path) {
     struct stat sb;
     if (stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)) return 1;
     return 0;
 }
 
-int is_numeric(const char * str) {
+static int is_numeric(const char * str) {
     for (; *str; str++)
         if (!isdigit(*str)) return 0;
     return 1;
 }
 
 int main(int argc, char * argv[]) {
-    int mode = MODE_UNSPECIFIED;
-    int args_status = 1;
+    int mode = MODE_ENCODE;
 
     // input and output file names
     char *input = NULL, *output = NULL;
-    char *bz3_file = NULL, *regular_file = NULL;
-    int no_bz3_suffix = 0, force = 0;
+    char *f1 = NULL, *f2 = NULL;
+    int force = 0;
 
     // command line arguments
     int force_stdstreams = 0, workers = 0;
     int double_dash = 0;
 
     // the block size
-    u32 block_size = MiB(8);
+    u32 block_size = MiB(16);
 
-    for (int i = 1; i < argc; i++) {
-        if (argv[i][0] == '-' && !double_dash) {
-            if (strlen(argv[i]) < 2) {
-                fprintf(stderr, "Invalid flag '%s'.\n", argv[i]);
-                return 1;
-            } else if (argv[i][1] == 'e') {
-                mode = MODE_ENCODE;
-            } else if (argv[i][1] == 'd') {
-                mode = MODE_DECODE;
-            } else if (argv[i][1] == 't') {
-                mode = MODE_TEST;
-            } else if (argv[i][1] == 'f') {
-                force = 1;
-            } else if (argv[i][1] == 'b') {
-                if (i + 1 >= argc) {
-                    fprintf(stderr, "Error: -b requires an argument.\n");
-                    return 1;
-                }
-
-                if (!is_numeric(argv[i + 1])) {
-                    fprintf(stderr, "Error: -b requires an integer argument.\n");
-                    return 1;
-                }
-
-                block_size = MiB(atoi(argv[i + 1]));
-                i++;
-            } else if (argv[i][1] == 'c') {
-                force_stdstreams = 1;
 #ifdef PTHREAD
-            } else if (argv[i][1] == 'j') {
-                if (i + 1 >= argc) {
-                    fprintf(stderr, "Error: -j requires an argument.\n");
-                    return 1;
-                }
+    const char * getopt_args = "b:cdefhj:tV";
+#else
+    const char * getopt_args = "b:cdefhtV";
+#endif
 
-                if (!is_numeric(argv[i + 1])) {
-                    fprintf(stderr, "Error: -j requires an integer argument.\n");
-                    return 1;
-                }
-
-                workers = atoi(argv[i + 1]);
-                i++;
+    operr = 1; // Should be set by default, just make sure.
+    while (opind < argc) {
+        int opt;
+        if((opt = getopt_impl(argc, argv, getopt_args)) != -1) {
+            // Normal dash argument.
+            switch(opt) {
+                case 'e':
+                    mode = MODE_ENCODE;
+                    break;
+                case 'd':
+                    mode = MODE_DECODE;
+                    break;
+                case 't':
+                    mode = MODE_TEST;
+                    break;
+                case 'f':
+                    force = 1;
+                    break;
+                case 'c':
+                    force_stdstreams = 1;
+                    break;
+                case 'V':
+                    version();
+                    return 0;
+                case 'h':
+                    help();
+                    return 0;
+                case 'b':
+                    if (is_numeric(oparg)) {
+                        block_size = MiB(atoi(oparg));
+                    } else {
+                        fprintf(stderr, "Invalid block size: %s\n", oparg);
+                        return 1;
+                    }
+                    break;
+#ifdef PTHREAD
+                case 'j':
+                    if (is_numeric(oparg)) {
+                        workers = atoi(oparg);
+                    } else {
+                        fprintf(stderr, "Invalid number of workers: %s\n", oparg);
+                        return 1;
+                    }
+                    break;
 #endif
-            } else if (argv[i][1] == 'h') {
-                mode = MODE_UNSPECIFIED;
-                args_status = 0;
-            } else if (argv[i][1] == 'v') {
-                fprintf(stderr, "bzip3 %s\n", VERSION);
-                return 0;
-            } else if (argv[i][1] == '-') {
-                double_dash = 1;
-            } else {
-                fprintf(stderr, "Invalid flag '%s'.\n", argv[i]);
-                return 1;
             }
         } else {
-            if (bz3_file != NULL && regular_file != NULL) {
+            // Positional argument. Likely a file name.
+            char * arg = argv[opind++];
+
+            if (f1 != NULL && f2 != NULL) {
                 fprintf(stderr, "Error: too many files specified.\n");
                 return 1;
             }
 
-            int has_bz3_suffix = strlen(argv[i]) > 4 && !strcmp(argv[i] + strlen(argv[i]) - 4, ".bz3");
+            // int has_bz3_suffix = strlen(arg) > 4 && !strcmp(arg + strlen(arg) - 4, ".bz3");
 
-            if (has_bz3_suffix || regular_file != NULL) {
-                bz3_file = argv[i];
-                no_bz3_suffix = !has_bz3_suffix;
-            } else {
-                regular_file = argv[i];
-            }
+            if (f1 == NULL)
+                f1 = arg;
+            else
+                f2 = arg;
         }
     }
-
-    if (mode == MODE_UNSPECIFIED) {
-        fprintf(stderr, "bzip3 version %s\n", VERSION);
-        fprintf(stderr, "- A better and stronger spiritual successor to bzip2.\n");
-        fprintf(stderr, "Copyright (C) by Kamila Szewczyk, 2022. Licensed under the terms of LGPLv3.\n");
-        fprintf(stderr, "Usage: bzip3 [-e/-d/-t/-c/-h/-v] [-b block_size] input output\n");
-        fprintf(stderr, "Operations:\n");
-        fprintf(stderr, "  -e: encode\n");
-        fprintf(stderr, "  -d: decode\n");
-        fprintf(stderr, "  -t: test\n");
-        fprintf(stderr, "  -h: help\n");
-        fprintf(stderr, "  -f: force overwrite output if it already exists\n");
-        fprintf(stderr, "  -v: version\n");
-        fprintf(stderr, "Extra flags:\n");
-        fprintf(stderr, "  -c: force reading/writing from standard streams\n");
-        fprintf(stderr, "  -b N: set block size in MiB\n");
-#ifdef PTHREAD
-        fprintf(stderr, "  -j N: set the amount of parallel threads\n");
-#endif
-        return args_status;
-    }
 #ifndef O_BINARY
     #define O_BINARY 0
 #endif
@@ -166,34 +231,47 @@ int main(int argc, char * argv[]) {
     setmode(STDOUT_FILENO, O_BINARY);
 #endif
 
-    if (mode != MODE_TEST) {
-        if (no_bz3_suffix || mode == MODE_ENCODE) {
-            input = regular_file;
-            output = bz3_file;
-            if (!force_stdstreams && !no_bz3_suffix && output == NULL && input != NULL) {
-                // add the bz3 extension
-                output = malloc(strlen(input) + 5);
-                strcpy(output, input);
-                strcat(output, ".bz3");
-            }
-        } else {
-            input = bz3_file;
-            output = regular_file;
-            if (!force_stdstreams && output == NULL && input != NULL) {
-                // strip the bz3 extension
-                if (strlen(input) > 4 && !strcmp(input + strlen(input) - 4, ".bz3")) {
-                    output = malloc(strlen(input));
-                    strncpy(output, input, strlen(input) - 4);
-                    output[strlen(input) - 4] = '\0';
+    if (mode == MODE_TEST)
+        input = f1;
+    else {
+        if (mode == MODE_ENCODE) {
+            if(f2 == NULL) {
+                // encode from f1.
+                input = f1;
+                if(force_stdstreams)
+                    output = NULL;
+                else {
+                    output = (char *)malloc(strlen(f1) + 5);
+                    strcpy(output, f1);
+                    strcat(output, ".bz3");
                 }
+            } else {
+                // encode from f1 to f2.
+                input = f1; output = f2;
+            }
+        } else if(mode == MODE_DECODE) {
+            if(f2 == NULL) {
+                // decode from f1 to stdout.
+                input = f1;
+                if(force_stdstreams)
+                    output = NULL;
+                else {
+                    output = (char *)malloc(strlen(f1) + 1);
+                    strcpy(output, f1);
+                    if(strlen(output) > 4 && !strcmp(output + strlen(output) - 4, ".bz3"))
+                        output[strlen(output) - 4] = 0;
+                    else {
+                        fprintf(stderr, "Warning: file %s has an unknown extension, skipping.\n", f1);
+                        return 1;
+                    }
+                }
+            } else {
+                // decode from f1 to f2.
+                input = f1; output = f2;
             }
         }
-    } else {
-        input = bz3_file != NULL ? bz3_file : regular_file;
     }
-
-    if (input == NULL && output == NULL) force_stdstreams = 1;
-
+    
     FILE *input_des = NULL, *output_des = NULL;
 
     if (input != NULL) {
@@ -320,6 +398,7 @@ int main(int argc, char * argv[]) {
                 fwrite(byteswap_buf, 4, 1, output_des);
                 fwrite(buffer, new_size, 1, output_des);
             }
+            fflush(output_des);
         } else if (mode == MODE_DECODE) {
             s32 new_size, old_size;
             while (!feof(input_des)) {
@@ -343,6 +422,7 @@ int main(int argc, char * argv[]) {
                 }
                 fwrite(buffer, old_size, 1, output_des);
             }
+            fflush(output_des);
         } else if (mode == MODE_TEST) {
             s32 new_size, old_size;
             while (!feof(input_des)) {
@@ -423,13 +503,17 @@ int main(int argc, char * argv[]) {
                     fwrite(buffers[j], sizes[j], 1, output_des);
                 }
             }
+            fflush(output_des);
         } else if (mode == MODE_DECODE) {
             while (!feof(input_des)) {
                 s32 i = 0;
                 for (; i < workers; i++) {
                     if (fread(&byteswap_buf, 1, 4, input_des) != 4) break;
                     sizes[i] = read_neutral_s32(byteswap_buf);
-                    if (fread(&byteswap_buf, 1, 4, input_des) != 4) break;
+                    if (fread(&byteswap_buf, 1, 4, input_des) != 4) {
+                        fprintf(stderr, "I/O error.\n");
+                        return 1;
+                    }
                     old_sizes[i] = read_neutral_s32(byteswap_buf);
                     if (fread(buffers[i], 1, sizes[i], input_des) != sizes[i]) {
                         fprintf(stderr, "I/O error.\n");
@@ -447,13 +531,17 @@ int main(int argc, char * argv[]) {
                     fwrite(buffers[j], old_sizes[j], 1, output_des);
                 }
             }
+            fflush(output_des);
         } else if (mode == MODE_TEST) {
             while (!feof(input_des)) {
                 s32 i = 0;
                 for (; i < workers; i++) {
                     if (fread(&byteswap_buf, 1, 4, input_des) != 4) break;
                     sizes[i] = read_neutral_s32(byteswap_buf);
-                    if (fread(&byteswap_buf, 1, 4, input_des) != 4) break;
+                    if (fread(&byteswap_buf, 1, 4, input_des) != 4) {
+                        fprintf(stderr, "I/O error.\n");
+                        return 1;
+                    }
                     old_sizes[i] = read_neutral_s32(byteswap_buf);
                     if (fread(buffers[i], 1, sizes[i], input_des) != sizes[i]) {
                         fprintf(stderr, "I/O error.\n");
